/* crystile.c */
#define VERSION "crystile 1.1, placed in public domain by Steve Kirkendall, 8 April 1996"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "tile.h"
extern char *optarg;
extern int optind;
extern char **readopt(FILE *fp);

typedef enum {I_ORIENT, I_AGE, I_SEED} itype_t;

/* These variables are set from command-line flags */
int	dither = 1;		/* perform random dithering? */
int	maxpalette = 6;		/* number of colors in spectrum */
int	bounce = 4;		/* number of distinct colors in palette */
int	cw, ccw;		/* travel direction when interpolating colors */
double	mustfreeze = 1.0;	/* fraction of points that must be frozen */
double	grain = 0.5;		/* probability of growing along grain */
int	curl = 0;		/* curl factor */
int	reversecurl = 0;	/* reverse curl factor */
char	*light = "lightblue";	/* first color in spectrum */
char	*dark = "darkblue";	/* last color in spectrum */
char	*thaw = "darkdarkblue";	/* color of the unfrozen areas */
char	outfile[255];		/* name of the output file */
int	width=128, height=128;	/* size of the image */
int	nseeds = 10;		/* number of seed points */
itype_t	itype = I_ORIENT;	/* image type */
int	agemult = 1;		/* number of generations per age */

/* These store the texture colors, as 8-bit RGB values */
rgb_t	lightrgb;
rgb_t	darkrgb;
rgb_t	thawrgb;

/* These store information about each pixel. */
unsigned char   *imagedata;
unsigned char   *olddata;
unsigned char	*orientation;
#define image(x,y)	(imagedata[(y) * width + (x)])
#define oldimage(x,y)	(olddata[(y) * width + (x)])
#define orient(x,y)	(orientation[(y) * width + (x)])
#define isfreezing(x,y)	(image((x),(y)) != maxpalette)
#define wasfrozen(x,y)	(oldimage((x),(y)) != maxpalette)
#define wasfrozeni(i)	(olddata[i] != maxpalette)
#define wrapx(x)	(((x) + width) % width)
#define wrapy(y)	(((y) + height) % height)

/* These store orientation-specific values */
unsigned char	ocolor[256];	/* color for a particular orientation */
double		oprob[256];	/* probability of dithering to next color */
double		odx[256];	/* x delta for grain */
double		ody[256];	/* y delta for grain */

/* function declarations */
void usage(char *hint);
void allocimg(void);
void main(int argc, char **argv);


void usage(char *hint)
{
#ifdef SUPPORT_X11
	fprintf(stderr, "usage: crystile [options] [outfile[.gif]]\n");
#else
	fprintf(stderr, "usage: crystile [options] outfile[.gif]\n");
#endif
	fprintf(stderr, "   -v           Display the version number; don't generate an image\n");
	fprintf(stderr, "   -r           Disable random dithering of \"orient\" image type\n");
	fprintf(stderr, "   -s [w,]h     Size of image; default is 128,128\n");
	fprintf(stderr, "   -i style     Image coloring style: orient/seed/age[n]; default orient\n");
	fprintf(stderr, "   -p [c,]s     Size of color palette, and optional number of colors in it\n");
	fprintf(stderr, "   -t inter     Interpolation path from dark to light color: cw/ccw/chord\n");
	fprintf(stderr, "   -g grain     Probability of growing along grain; default 0.5\n");
	fprintf(stderr, "   -c curl[,c2] Curling factors, 0~127; default 0 and same\n");
	fprintf(stderr, "   -n seeds     Number of seeds\n");
	fprintf(stderr, "   -l light     Name of light color; default is lightblue\n");
	fprintf(stderr, "   -d dark      Name of dark color; default is darkblue\n");
	fprintf(stderr, "   -w thawed    Name of unfrozen color (for -f#); default is white\n");
	fprintf(stderr, "   -f freeze    Fraction of pixels to freeze; default is 1.0\n");
	fprintf(stderr, "   -m macro     Use options from a line in crystile.mac or ~/.crystile\n");
	fprintf(stderr, "   -M macro     Display options from a line in crystile.mac or ~/.crystile\n");
	if (hint)
	{
		fprintf(stderr, "%s\n", hint);
	}
	else
	{
		fprintf(stderr, "This program uses a random number generator to produce background image tiles\n");
		fprintf(stderr, "for use with Netscape or other graphical utilities.  It simulates the growth\n");
		fprintf(stderr, "of colored crystals.  The output file can use either Compuserve's GIF image\n");
#ifdef SUPPORT_X11
		fprintf(stderr, "file format, or Microsoft's BMP format.  If no output file name is specified,\n");
		fprintf(stderr, "this program will attempt to set the background image of your X terminal.\n");
#else
		fprintf(stderr, "file format, or Microsoft's BMP format.\n");
#endif
	}
	exit(1);
}




/* Allocate memory for the color and orientation of each pixel */
void allocimg(void)
{
	/* if we would need to allocate too much RAM for a size_t, complain */
	if ((long)width * (long)height >= 65536 && sizeof(size_t) == 2)
	{
		fprintf(stderr, "This computer can only generate 257x255 images\n");
		exit(2);
	}

	/* allocate the imagedata and orientation arrays */
	imagedata = (unsigned char *)malloc(width * height * sizeof(unsigned char));
	olddata = (unsigned char *)malloc(width * height * sizeof(unsigned char));
	orientation = (unsigned char *)malloc(width * height * sizeof(unsigned char));

	/* initially, load the imagedata with the thaw color */
	memset(imagedata, maxpalette, width * height);
}



/* Try to locate a frozen neighbor.  If there is none, then return 0; else
 * set *nx and *ny to its coordinates and return 1.
 */
int frozenneighbor(int x, int y, int *nx, int *ny)
{
	static int dx[8] = { 0,  1,  1,  1,  0, -1, -1, -1};
	static int dy[8] = {-1, -1,  0,  1,  1,  1,  0, -1};
	int frozenx[8], frozeny[8];
	int	i, qty;

	/* count the neighbors, and remember where they are */
	qty = 0;
	if (x == 0 || y == 0 || x == width - 1 || y == height - 1)
	{
		/* on the edge - be careful about wrapping */
		for (i = 0; i < 8; i++)
		{
			frozenx[qty] = wrapx(x + dx[i]);
			frozeny[qty] = wrapy(y + dy[i]);
			if (wasfrozen(frozenx[qty], frozeny[qty]))
				qty++;
		}
	}
	else
	{
		/* near center of tile, we don't need to worry about wrapping */
		i = y * width + x;
		if (wasfrozeni(i - width - 1))
			frozenx[qty] = x - 1, frozeny[qty] = y - 1, qty++;
		if (wasfrozeni(i - width))
			frozenx[qty] = x, frozeny[qty] = y - 1, qty++;
		if (wasfrozeni(i - width + 1))
			frozenx[qty] = x + 1, frozeny[qty] = y - 1, qty++;
		if (wasfrozeni(i - 1))
			frozenx[qty] = x - 1, frozeny[qty] = y, qty++;
		if (wasfrozeni(i + 1))
			frozenx[qty] = x + 1, frozeny[qty] = y, qty++;
		if (wasfrozeni(i + width - 1))
			frozenx[qty] = x - 1, frozeny[qty] = y + 1, qty++;
		if (wasfrozeni(i + width))
			frozenx[qty] = x, frozeny[qty] = y + 1, qty++;
		if (wasfrozeni(i + width + 1))
			frozenx[qty] = x + 1, frozeny[qty] = y + 1, qty++;
	}

	/* if none, then return 0 */
	if (qty == 0)
		return 0;

	/* choose one, ignore the others */
	if (qty == 1)
		i = 0;
	else
		i = drand48() * qty;
	*nx = frozenx[i];
	*ny = frozeny[i];
	return 1;
}


/* stuff the seed pixels into the image */
void putseeds()
{
	int	x, y;
	int	o;
	int	i;

	/* for each seed... */
	for (i = 0; i < nseeds; i++)
	{
		/* choose a random location.  Don't worry about overwriting
		 * a pixel that's already a seed.
		 */
		x = drand48() * width;
		y = drand48() * height;

		/* choose an orientation and color for this pixel */
		orient(x, y) = drand48() * 256;
		switch (itype)
		{
		  case I_ORIENT:
			o = orient(x,y);
			if (oprob[o] < 0.01 || drand48() > oprob[o])
				image(x, y) = ocolor[o];
			else if (ocolor[o] > 0)
				image(x,y) = ocolor[o] - 1;
			else
				image(x,y) = maxpalette - 1;
			break;

		  case I_AGE:
			image(x, y) = maxpalette - 1;
			break;

		  case I_SEED:
			image(x, y) = i * bounce / nseeds;
			break;
		}
	}
}


/* This function performs a single crystal growth pass, and returns the number
 * of pixels that are frozen.
 */
unsigned int growcrystal(int age)
{
	int	i, j;
	int	x, y;	/* a point in the image */
	int	o;	/* orientation of that point */
	int	nx, ny;	/* neighboring point */
	int	gx, gy;	/* a point along its grain */
	double	dgx,dgy;/* a point along its grain */
	int	go;	/* orientation of a a point along the grain */
	int	nfrozen;/* number of pixels that have frozen so far */

	/* Copy the image.  Pixels which freeze in this pass should not count
	 * as a "frozen neighbor" for the remainder of this pass, or else weird
	 * side-effects occur.  So we copy the image now and only look for
	 * frozen neighbors there.
	 */
	memcpy(olddata, imagedata, width * height);

	/* for each pixel... */
	for (y = 0; y < height; y++)
	{
		for (x = 0; x < width; x++)
		{
			/* skip if already frozen */
			if (isfreezing(x, y))
				continue;

			/* if no frozen neighbors, then skip it */
			if (!frozenneighbor(x, y, &nx, &ny))
				continue;

			/* freeze this point */
			orient(x, y) = orient(nx, ny);
			switch (itype)
			{
			  case I_ORIENT:
				o = orient(x,y);
				if (oprob[o] < 0.01 || drand48() > oprob[o])
					image(x, y) = ocolor[o];
				else if (ocolor[o] > 0)
					image(x,y) = ocolor[o] - 1;
				else
					image(x,y) = maxpalette - 1;
				break;

			  case I_AGE:
				image(x, y) = age;
				break;

			  case I_SEED:
				image(x, y) = image(nx, ny);
				break;
			}

			/* if grain probability is 0, we're done */
			if (grain < 0.01)
				continue;

			/* grow along the grain, in the forward direction */
			go = (orient(x, y) + curl) & 0xff;
			dgx = (double)x + 0.5 + odx[go];
			if (dgx < 0.0) dgx += (double)width;
			dgy = (double)y + 0.5 + ody[go];
			if (dgy < 0.0) dgy += (double)height;
			gx = wrapx((int)dgx);
			gy = wrapy((int)dgy);
			while (!isfreezing(gx, gy) && grain > drand48())
			{
				/* freeze this point, too */
				orient(gx, gy) = go;
				switch (itype)
				{
				  case I_ORIENT:
					if (oprob[go] < 0.01 || drand48() > oprob[go])
						image(gx, gy) = ocolor[go];
					else if (ocolor[go] > 0)
						image(gx,gy) = ocolor[go] - 1;
					else
						image(gx,gy) = maxpalette - 1;
					break;

				  case I_AGE:
					image(gx, gy) = age;
					break;

				  case I_SEED:
					image(gx, gy) = image(nx, ny);
					break;
				}

				/* locate the next point */
				go = (go + curl) & 0xff;
				dgx += odx[go];
				if (dgx < 0.0) dgx += (double)width;
				dgy += ody[go];
				if (dgy < 0.0) dgy += (double)height;
				gx = wrapx((int)dgx);
				gy = wrapy((int)dgy);
			}

			/* grow along the grain, in the reverse direction */
			go = (orient(x, y) + 0x100 - curl) & 0xff;
			dgx = (double)x + 0.5 - odx[go];
			if (dgx < 0.0) dgx += (double)width;
			dgy = (double)y + 0.5 - ody[go];
			if (dgy < 0.0) dgy += (double)height;
			gx = wrapx((int)dgx);
			gy = wrapy((int)dgy);
			while (!isfreezing(gx, gy) && grain > drand48())
			{
				/* freeze this point, too */
				orient(gx, gy) = go;
				switch (itype)
				{
				  case I_ORIENT:
					if (oprob[go] < 0.01 || drand48() > oprob[go])
						image(gx, gy) = ocolor[go];
					else if (ocolor[go] > 0)
						image(gx,gy) = ocolor[go] - 1;
					else
						image(gx,gy) = maxpalette - 1;
					break;

				  case I_AGE:
					image(gx, gy) = age;
					break;

				  case I_SEED:
					image(gx, gy) = image(nx, ny);
					break;
				}

				/* locate the next point */
				go = (go + 0x100 - reversecurl) & 0xff;
				dgx -= odx[go];
				if (dgx < 0.0) dgx += (double)width;
				dgy -= ody[go];
				if (dgy < 0.0) dgy += (double)height;
				gx = wrapx((int)dgx);
				gy = wrapy((int)dgy);
			}
		}
	}

	/* count how many are frozen now */
	nfrozen = 0;
	for (i = width * height; --i >= 0; )
		if (imagedata[i] != maxpalette)
				nfrozen++;

	return nfrozen;
}



void main(int argc, char **argv)
{
	int	opt;
	int	i, j;
	int	didm;
	char	macrofile[200];
	char	**macroargv;
	FILE	*fp;

	/* seed the random number generator */
	srand48(time(NULL) + getpid());

	/* convert the default color names to rgb_t values */
	thawrgb = name_to_RGB(thaw);
	lightrgb = name_to_RGB(light);
	darkrgb = name_to_RGB(dark);

	didm = 0;
	while ((opt = getopt(argc, argv, "vrs:i:p:t:g:c:n:s:f:l:d:w:m:M:?")) != EOF)
	{
		switch (opt)
		{
		  case 'v':
			puts(VERSION);
			exit(0);

		  case 'r':
		  	dither = 0;
		  	break;

		  case 's':
			if (sscanf(optarg, "%d,%d", &width, &height) == 2)
			{
			}
			else if (sscanf(optarg, "%d", &height) == 1)
			{
				width = height;
			}
			break;

		  case 'i':
			if (!strcmp(optarg, "random"))
			{
				switch ((int)(drand48() * 4))
				{
				  case 0:   itype = I_ORIENT;		break;
				  case 1:   itype = I_AGE;		break;
				  case 2:   itype = I_AGE, agemult = 4;	break;
				  default:  itype = I_SEED;
				}
			}
			else if (!strncmp(optarg, "orient", 6))
			{
				itype = I_ORIENT;
			}
			else if (!strncmp(optarg, "age", 3))
			{
				itype = I_AGE;
				if (optarg[3])
				{
					agemult = atoi(&optarg[3]);
					if (agemult < 1)
						usage("invalid age multiplier -- should be >= 1");
				}
			}
			else if (!strcmp(optarg, "seed"))
			{
				itype = I_SEED;
			}
			else
			{
				usage("image type must be one of {orient, age[n], seed, random");
			}
			break;

		  case 'p':
			if (sscanf(optarg, "%d,%d", &bounce, &maxpalette) != 2)
			{
				bounce = atoi(optarg);
				maxpalette = 2 * (bounce - 1);
			}
			if (bounce < 2)
				usage("must use at least two colors");
			if (maxpalette > 255)
				usage("can't use that many colors");
			if (bounce > maxpalette)
				usage("number of distinct colors must be smaller than palette size");
			break;

		  case 't':
			if (!strcmp(optarg, "cw"))
				cw = 1, ccw = 0;
			else if (!strcmp(optarg, "ccw"))
				cw = 0, ccw = 1;
			else if (!strcmp(optarg, "chord"))
				cw = ccw = 0;
			else if (!strcmp(optarg, "random"))
				switch ((int)(drand48() * 3))
				{
				  case 0:	cw = 1, ccw = 0;	break;
				  case 1:	cw = 0, ccw = 1;	break;
				  default:	cw = 0, ccw = 0;	break;
				}
			else
				usage("travel must be one of {cw, ccw, chord, random}");
			break;

		  case 'g':
			grain = atof(optarg);
			if (grain < 0.0 || grain > 1.0)
				usage("grain probability must be between 0.0 and 1.0");
			break;

		  case 'c':
			if (sscanf(optarg, "%d,%d", &curl, &reversecurl) != 2)
			{
				curl = atoi(optarg);
				reversecurl = curl;
			}
		  	if (curl < 0 || curl > 128 || abs(reversecurl) > 128)
		  		usage("curl factor must be between 0 and 128\n");
		  	break;

		  case 'n':
			nseeds = atoi(optarg);
			if (nseeds <= 0)
				usage("number of seeds must be at least 1\n");
			break;

		  case 'f':
			mustfreeze = atof(optarg);
			if (mustfreeze < 0.0 || mustfreeze > 1.0)
				usage("freezing fraction must be between 0.0 and 1.0");
			break;

		  case 'w':
		  	thaw = optarg;
			thawrgb = name_to_RGB(thaw);
			break;

		  case 'l':
			light = optarg;
			lightrgb = name_to_RGB(light);
			break;

		  case 'd':
			dark = optarg;
			darkrgb = name_to_RGB(dark);
			break;

		  case 'm':
		  case 'M':
			/* use of static variables prohibits nested macros */
			if (didm)
				usage("-mmacro can only occur once");
			didm = 1;

			/* open the macro file.  It may be "crystile.mac" in
			 * the current directory, ".crystile" in the HOME
			 * directory, or "crystile.mac" in the directory
			 * where the executable resides.
			 */
			fp = fopen("crystile.mac", "r");
			if (!fp)
			{
				if (getenv("HOME"))
				{
					sprintf(macrofile, "%s/.crystile", getenv("HOME"));
				}
				else if (strrchr(argv[0], '\\'))
				{
					strcpy(macrofile, argv[0]);
					strcpy(strrchr(argv[0], '\\') + 1, "crystile.mac");
				}
				else
				{
					perror("crystile.mac");
					exit(2);
				}
				fp = fopen(macrofile, "r");
				if (!fp)
				{
					perror(macrofile);
					exit(2);
				}
			}

			/* search for the macro */
			if (!strcmp(optarg, "random"))
			{
				/* count the number of macros */
				for (i = 0; readopt(fp) != NULL; i++)
				{
				}
				rewind(fp);

				/* if empty file, then complain */
				if (i == 0)
				{
					fprintf(stderr, "%s: no macros defined\n", argv[0]);
					exit(2);
				}

				/* choose macro randomly */
				i *= drand48();

				/* read that macro */
				do
				{
					macroargv = readopt(fp);
				} while (--i >= 0);
				fclose(fp);
			}
			else
			{
				/* look for a specific name */
				while ((macroargv = readopt(fp)) != NULL &&
					(!macroargv[0] || strcmp(macroargv[0], optarg)))
				{
				}
				fclose(fp);
				if (macroargv == NULL)
				{
					fprintf(stderr, "%s: macro \"%s\" not found\n", argv[0], optarg);
					exit(2);
				}
			}

			if (opt == 'M')
			{
				/* display the macro name and its options */
				for (i = 0; macroargv[i]; i++)
				{
					printf("%s%s", i>0?" ":"", macroargv[i]);
				}
				printf("\n");
				exit(0);
			}
			else if (macroargv[1] != NULL)
			{
				/* skip the macro name; use its options */
				putopt(1, macroargv);
			}
			break;

		  default:
			usage(NULL);
		}
	}

	/* find the filename.  If no filename, then give usage message */
	if (optind + 1 == argc)
	{
		strcpy(outfile, argv[optind]);
		if (strchr(outfile, '.') == NULL)
			strcat(outfile, ".gif");
	}
#ifdef SUPPORT_X11
	else if (optind != argc)
	{
		usage("you can only specify one output name, or none");
	}
#else
	else
	{
		usage("you must specify one output file name");
	}
#endif

	/* some sanity checks */
	if (height < 10 || width < 10 || width * height < nseeds * 10)
		usage("image would be too small");

	/* generate a color palette */
	genpalette(darkrgb, lightrgb, thawrgb, maxpalette, cw, ccw, bounce);

	/* compute some tables */
	for (i = 0; i < 256; i++)
	{
		ocolor[i] = maxpalette - 1 - (maxpalette * i / 256);
		if (dither)
			oprob[i] = (double)(maxpalette * i % 256) / 256.0;
		odx[i] = -cos((double)i * M_PI / 128.0);
		ody[i] = -sin((double)i * M_PI / 128.0);
	}

	/* allocate the pixel arrays */
	allocimg();

	/* stuff the seed points into the image */
	putseeds();

	/* grow the crystal until the required number of points are frozen */
	opt = (int)((width * height) * mustfreeze);
	j = agemult;
	for (i = maxpalette - 1; growcrystal(i) < opt; )
	{
		if (--j == 0)
		{
			j = agemult;
			i--;
			if (i < 0)
				i = maxpalette - 1;
		}
	}

	/* reduce the color palette by eliminating duplicate/unused colors */
	reduce(imagedata, width, height);
	
	/* output the image */
#ifdef SUPPORT_X11
	if (*outfile == '\0')
		WriteX11(outfile, imagedata, width, height, palette, palettesize);
	else
#endif
	if ((outfile[strlen(outfile) - 1] & 0x5f) == 'P')
		WriteBmp(outfile, imagedata, width, height, palette, palettesize);
	else
		WriteGif(outfile, imagedata, width, height, palette, palettesize);

	exit(0);
}
